// QDSManager.cpp: implementation of the QDSManager class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Query32.h"
#include "QDSManager.h"

extern "C" {
#ifdef _NS_LDAP
#include "ldap.h"
#else
#include "winldap.h"
#endif
}

#include <limits.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

QDSManager::QDSManager()
{
	WIN32_GUFF();
  m_nReqIDAcc = 0;
  m_bPolling = FALSE;
	m_pFormConfig = new QFormConfiguration( this );
}

QDSManager::~QDSManager()
{
  WIN32_GUFF();

  POSITION pos = m_requestList.GetHeadPosition();
  while ( pos != NULL )
    {
      QRequest *request = m_requestList.GetNext( pos );
			request->Abort();
      delete request;
    }

	m_requestList.RemoveAll();
	delete m_pFormConfig;

	//
	// Close log file
	QLogFile::Close();
}

//////////////////////////////////////////////////////////////////////
// QDSManager::Waiting
// Check to see if there are any requests outstanding.
BOOL QDSManager::Waiting()
{
  WIN32_GUFF();

	if ( m_requestList.GetCount() == 0 )
		return FALSE;

	int nWaitCount = 0;
	POSITION pos = m_requestList.GetHeadPosition();
	while ( pos != NULL )
		{
			QRequest *req = m_requestList.GetNext( pos );
			if ( req->Finished() == FALSE )
				nWaitCount++;
		}

	return ( nWaitCount > 0? TRUE : FALSE );
}

//////////////////////////////////////////////////////////////////////
// QDSManager::WaitingOn
// Check to see if there are any requests outstanding on the 
// ID'ed connection.
BOOL QDSManager::WaitingOn( ConnID conn_id )
{
  WIN32_GUFF();

	if ( m_requestList.GetCount() == 0 )
		return FALSE;

	POSITION pos = m_requestList.GetHeadPosition();
	while ( pos != NULL )
		{
			QRequest *req = m_requestList.GetNext( pos );
			if ( req->WaitingOn( conn_id ) == TRUE )
				return TRUE;
		}

	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// QDSManager::Poll
// Poll the all directory connections to handle returned LDAP
// result messages and errors.
void QDSManager::Poll()
{
  WIN32_GUFF();

  // No work to do?
  if ( m_requestList.GetCount() == 0 )
    return;

  // Make sure that we don't re-enter.
  if ( m_bPolling == TRUE )
    return;

  m_bPolling = TRUE;

  // All done
  struct timeval timeout;
  BOOLEAN done = FALSE;
  int lerr;
  LDAPMessage *lm;

  timeout.tv_sec = 0;
  timeout.tv_usec = 0;

  // Poll on each open connection
  POSITION pos = m_bindMgr.GetConnectionMap()->GetHeadPosition();
  while ( pos != NULL )
    {
      QConnection *conn;
      
      // Get the next connection
      conn = m_bindMgr.GetConnectionMap()->GetNext( pos );
			if ( conn->Bound() == FALSE )
				continue;

			//
      // Continue m_bPolling on this connection until there is no more pending data
			if ( WaitingOn( conn->ID() ) == FALSE )
				continue;

      do
        {
					//
					// Set the activity timestamp
					conn->UpdateLastActive();

					//
					// Look for an LDAP result
          lerr = ldap_result
						( conn->GetLDAPContext(), LDAP_RES_ANY, LDAP_MSG_ONE, &timeout, &lm );

					if ( lerr == -1 )
						//
						// An error occured
						{
							char *pMatched;
							char *pErr;

							int nErr =
								ldap_get_lderrno( conn->GetLDAPContext(), &pMatched, &pErr );

							QLOG( "QDSManager::Poll()" );
							QLOGRAW( "ldap_result returned an error:" );
							QLOGRAW( "Error Code: %d\nError String: %s", nErr, pErr );

							//
							// Take action if the connection has gone down
							if ( nErr == LDAP_SERVER_DOWN )
								{
									this->NotifyConnectionClose( conn->ID() );
								}

							//
							// Stop m_bPolling on this connection
							break;
						}
					
          if ( lerr == 0 )
						//
	          // No responses on this connection
	          {
	            break;
	          }

					//
					// If things get this far then a response has been received
					QRequest *req = 
						GetOpOwner( conn->GetLDAPContext(), ldap_msgid( lm ) );

					QError *err = NULL;
					if ( req != NULL )
	 					req->HandleLDAPMessage( conn->GetLDAPContext(), lm, err );

					//
					// XXX Error messages from QRequest::HandleLDAPMessage ignored in QDSManager::Poll
					if ( err != NULL )
						{
							delete err;
							err = NULL;
						}

					//
					// If the request has completed take it off the active list
					if ( req != NULL && req->Finished() == TRUE )
						DeregisterRequest( req->ID() );

					//
					// Done with this LDAP result message
					ldap_msgfree( lm );
        }
			while
				( m_requestList.GetCount() > 0 &&
						WaitingOn( conn->ID() ) == TRUE &&
						conn->SelectRead( &timeout ) > 0 );
    }

	m_bPolling = FALSE;
}

//////////////////////////////////////////////////////////////////////
// QDSManager::RegisterRequest
// Register a newly invoked request.
void QDSManager::RegisterRequest( QRequest *request )
{
  WIN32_GUFF();
  m_requestList.AddTail( request );
}

//////////////////////////////////////////////////////////////////////
// QDSManager::DeregisterRequest
// Notify of request completion
void QDSManager::DeregisterRequest( QRequestID req_id )
{
  WIN32_GUFF();

  POSITION pos = m_requestList.GetHeadPosition();
  while ( pos != NULL )
    {
			POSITION last_pos = pos;
      QRequest *req = m_requestList.GetNext( pos );
      if ( req->ID() == req_id )
        {
          m_requestList.RemoveAt( last_pos );
        }
    }
}

//////////////////////////////////////////////////////////////////////
// QDSManager::GenRequestID
// Generate a request record identifier
QRequestID QDSManager::GenRequestID()
{
	WIN32_GUFF();

	if ( m_nReqIDAcc == ULONG_MAX )
		return ( m_nReqIDAcc = 0 );
	else
		{
			QRequestID ret = m_nReqIDAcc++;
			return ( ret );
		}
}

//////////////////////////////////////////////////////////////////////
// QDSManager::GenRequestID
// Get the identified request record
QRequest *QDSManager::GetRequest( QRequestID req_id )
{
  WIN32_GUFF();

  POSITION pos = m_requestList.GetHeadPosition();
  while ( pos != NULL )
    {
      QRequest *req = m_requestList.GetNext( pos );
      if (req->ID() == req_id)
        return req;
    }

  return NULL;
}

//////////////////////////////////////////////////////////////////////
// QDSManager::GetOpOwner
// Get the request that issued the identified LDAP request.
QRequest *QDSManager::GetOpOwner( LDAP *ld, int ldap_msgid )
{
  WIN32_GUFF();

  POSITION pos = m_requestList.GetHeadPosition();
  while ( pos != NULL )
    {
      QRequest *req = m_requestList.GetNext( pos );
      if ( req->IsTaskOwner( ld, ldap_msgid ) == TRUE )
        return req;
    }

  return NULL;
}

//////////////////////////////////////////////////////////////////////
// QDSManager::AbortRequest
// Abort the identified request.
void QDSManager::AbortRequest(QRequestID req_id)
{
  WIN32_GUFF();

  POSITION pos = m_requestList.GetHeadPosition();
  while ( pos != NULL )
    {
      QRequest *req = m_requestList.GetNext( pos );
      if (req->ID() == req_id)
        {
          m_requestList.RemoveAt( pos );
          req->Abort();
          delete req;
        }
    }
}

//////////////////////////////////////////////////////////////////////
// QDSManager::AbortAll
// Abort all outstanding requests.
void QDSManager::AbortAll()
{
  WIN32_GUFF();

  if ( m_requestList.GetCount() == 0 )
    return;

  POSITION pos = m_requestList.GetHeadPosition();
  while ( pos != NULL )
    {
      QRequest *req = m_requestList.GetNext( pos );
      req->Abort();
      delete req;
    }

  m_requestList.RemoveAll();
}

//////////////////////////////////////////////////////////////////////
// QDSManager::PreInitialise()
// Load OID table.
BOOL QDSManager::PreInitialise()
{
	WIN32_GUFF();

	//
	// Load the oidtable
  if ( m_oidtable.Load( "oidtable32" ) == FALSE )
		return FALSE;

	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// QDSManager::Initialise
// Load form configuration.
BOOL QDSManager::Initialise( CString &config_dn, CString &host, int port )
{
	WIN32_GUFF();

  if ( m_pFormConfig->LoadConfiguration( config_dn, host, port ) == FALSE )
		return FALSE;

  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// QDSManager::Open
// Open the log file.
void QDSManager::Open( const char *pFileName )
{
	WIN32_GUFF();
	QLogFile::Open( pFileName );
}

//////////////////////////////////////////////////////////////////////
// QDSManager::Log
// Write a message to the log file.
void QDSManager::Log( const char *pFormatStr, ... )
{
	WIN32_GUFF();
	va_list argList;
	va_start( argList, pFormatStr );
	QLogFile::LogV( pFormatStr, argList );
	va_end( argList );
}

//////////////////////////////////////////////////////////////////////
// QDSManager::NotifySocketClose
// Notify of a close event on the given socket.
void QDSManager::NotifySocketClose( SOCKET socket )
{
	WIN32_GUFF();

	//
	// Notify the bind manager of socket close event.
	ConnID connID = BindManager()->NotifySocketClose( socket );

	//
	// Notify requests of socket close event.
	if ( connID != UNDEF_CONN_ID )
		{
			if ( m_requestList.GetCount() == 0 )
				return;

			POSITION pos = m_requestList.GetHeadPosition();
			while ( pos != NULL )
				{
					QRequest *req = m_requestList.GetNext( pos );
					if ( req->WaitingOn( connID ) == TRUE )
						{
							req->NotifyCloseConnection( connID );
						}
				}
		}
}

//////////////////////////////////////////////////////////////////////
// QDSManager::NotifyConnectionClose
// Notify of connection termination.
void QDSManager::NotifyConnectionClose( ConnID conn_id )
{
	WIN32_GUFF();

	//
	// Notify requests of connection termination.
	if ( m_requestList.GetCount() == 0 )
		return;

	POSITION pos = m_requestList.GetHeadPosition();
	while ( pos != NULL )
		{
			QRequest *req = m_requestList.GetNext( pos );
			if ( req->WaitingOn( conn_id  ) == TRUE )
				{
					req->NotifyCloseConnection( conn_id );
				}
		}

	//
	// Notify the connection manager
	BindManager()->Unbind( conn_id );
}

//////////////////////////////////////////////////////////////////////
// QDSManager::TidyUp
// Do any housekeeping. At present this just calls the same named
// procedure in QBindManager (closes idle timed out DSA connections).
void QDSManager::TidyUp()
{
	if ( Waiting() == FALSE )
		//
		// Only tidy connections if there is no work outstanding
		BindManager()->TidyUp();
}
