// QReadEntry.cpp: implementation of the QReadEntry class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Query32.h"
#include "QDSManager.h"
#include "QReadEntry.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

QReadEntry::QReadEntry
	( QDSManager *pManager, QRequestID nReqID, QRequester *pRequester, QDN &rDN )
	:
QRequest( pManager, nReqID, pRequester )
{
	m_DN = rDN;
	m_nMsgID = -1;
}

QReadEntry::~QReadEntry()
{
	if ( Invoked() == TRUE && Finished() == TRUE )
		{
			Abort();
			Notify( DUAabort, NULL );
		}
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::Invoke
//
// Invoke the read operation.
BOOL QReadEntry::Invoke( QError *&rErr )
{
	rErr = NULL;

	//
	// Make sure invokes don't happen when a request is outstanding
	if ( invoked == TRUE && finished == FALSE )
		{
			rErr = new QError( QERR_SVC_BUSY );
			return FALSE;
		}

	//
	// Initiate a bind if the connection isn't open
	BOOL mBinding = FALSE;
	if ( m_DN.DS()->BindManager()->Bound( m_DN.GetConnID() ) == FALSE )
		{
			if ( m_DN.DS()->BindManager()->BindAsync( m_DN.GetConnID(), rErr ) == FALSE )
				return FALSE;
		}

	//
	// Get search base DN in string form
	CString dnStr;
	m_DN.GetDN( dnStr );

	//
	// Search scope
	int scope = LDAP_SCOPE_BASE;

	//
	// No timeout for the search operation
	struct timeval zerotime;
	zerotime.tv_sec = zerotime.tv_usec = 0L;
	
	//
	// Wait for the connection to come alive
	if ( m_DN.DS()->BindManager()->BindWait( m_DN.GetConnID(), rErr ) == FALSE )
		return FALSE;

	//
	// Invoke LDAP search operation
	char *filter = "(objectClass=*)";
	int nLdapErr = ldap_search_ext
		( GetLDAPContext(),
			(LPCTSTR) dnStr, scope, filter, NULL, 0,
			NULL, NULL, NULL, -1, &m_nMsgID );

	//
	// Get the connection details for logging purposes
	CString host, user, password;
	int port;
	m_DN.DS()->BindManager()->GetConnection( m_DN.GetConnID(), host, port, user, password );

	//
	// Check for errors
	BOOL bOK;
	if ( nLdapErr != LDAP_SUCCESS )
		{
			CString errStr;
			rErr = new QError( nLdapErr );
			rErr->Print( errStr );

			//
			// Log the error
			QLOG( "QReadEntry::Invoke" );
			QLOGRAW( "LDAP search (read) invocation failed:" );
			QLOGRAW( "LDAP Error: %s", errStr );
			QLOGRAW( "DN: %s", dnStr );
			QLOGRAW( "Filter: %s", filter );
			QLOGRAW( "Host: %s", host );

			//
			// Send an error report
			QReadErrorReport *pReport = new QReadErrorReport();
			pReport->Error() = *rErr;
			pReport->DN() = m_DN;
			pReport->Host() = host;
			Notify( DUAerror, pReport );

			bOK = FALSE;
		}
	else
		{
			//
			// Log the invocation
			QLOG( "QReadEntry::Invoke" );
			QLOGRAW( "LDAP search (read) invoked OK:" );
			QLOGRAW( "Message ID: %d", m_nMsgID );
			QLOGRAW( "DN: %s", dnStr );
			QLOGRAW( "Filter: %s", filter );
			QLOGRAW( "Host: %s", host );

			//
			// Send a user report
			QReadReport *pReport = new QReadReport( QRRInvoked );
			pReport->DN() = m_DN;
			pReport->Host() = host;
			Notify( DUAreport, pReport );

			//
			// Update state
			invoked = TRUE;
			finished = FALSE;

			bOK = TRUE;
			Manager()->RegisterRequest( this );
		}


	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::HandleLDAPMessage
//
// Handle the LDAP result message.
BOOL QReadEntry::HandleLDAPMessage( LDAP *pLDAP, LDAPMessage *pLDAPMsg, QError *&rErr )
{
	BOOL bOk;

	switch ( ldap_msgtype( pLDAPMsg ) )
		{
		case LDAP_RES_SEARCH_ENTRY:
			bOk = ParseEntry( pLDAP, pLDAPMsg, rErr );
			break;

		case LDAP_RES_SEARCH_REFERENCE:
			bOk = ParseReferral( pLDAP, pLDAPMsg ,rErr );
			break;

		case LDAP_RES_SEARCH_RESULT:
			bOk = ParseResult( pLDAP, pLDAPMsg, rErr );
			break;
		}

	return bOk;
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::ParseEntry
//
// Parse an LDAP entry result message.
BOOL QReadEntry::ParseEntry( LDAP *pLDAP, LDAPMessage *pLDAPMsg, QError *&rErr )
{
	m_pResultEntry = new QEntry( Manager(), m_DN.GetConnID(), pLDAP, pLDAPMsg );
  if ( m_pResultEntry->IsSet() == TRUE )
    {
			QReadResult *pResult = new QReadResult( m_pResultEntry );
			m_pResultEntry = NULL;

			Notify( DUAresult, static_cast< QReadResult * >( pResult ) );
      return TRUE;
    }
  else
    {
      delete m_pResultEntry;
			m_pResultEntry = NULL;
      return FALSE;
    }
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::ParseReferral XXX Unimplemented
//
// Parse an LDAP referral message.
BOOL QReadEntry::ParseReferral( LDAP *pLDAP, LDAPMessage *pLDAPMsg, QError *&rErr )
{
	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::ParseResult XX Referrals not handled
//
// Parse an LDAP result message.
BOOL QReadEntry::ParseResult( LDAP *pLDAP, LDAPMessage *pLDAPMsg, QError *&rErr )
{
  LDAPControl **serverctrls;
  char *matched_msg = NULL, *error_msg = NULL, **referrals;

  int rc;
  int parse_rc = ldap_parse_result
    ( GetLDAPContext(), pLDAPMsg, &rc, &matched_msg, &error_msg, &referrals, &serverctrls, 0 );

  if ( parse_rc != LDAP_SUCCESS )
    {
			QLogFile::Log
				( "QReadEntry::ParseResult()\
					\nFailed to parse LDAP result:\
					\nMessage ID: %s",
					ldap_msgid( pLDAPMsg ) );

      return FALSE;
    }

  // Check the results of the LDAP operation.
	if ( rc == LDAP_REFERRAL )
		{
			char *ref_str = strdup( referrals[0] );
			LDAPURLDesc *ludp = NULL;

			int ref_ok = ldap_url_parse( ref_str, &ludp );
			if ( ref_ok == LDAP_SUCCESS )
				{
					CString full_ref_str;
					CString dn_str;
					CString scope_str;

					//
					// Get DN and seacrh scope string
					m_DN.GetDN( dn_str );
					scope_str = ( ludp->lud_scope == LDAP_SCOPE_SUBTREE ? "sub" : "one" );

					//
					// Format the referral string
					full_ref_str.Format
						( "ldap://%s:%d/%s?objectClass?%s?%s",
							ludp->lud_host,
							ludp->lud_port,
							ludp->lud_dn,
							( LPCTSTR ) scope_str,
							"(objectclass=*)" );

					QLOG( "QReadEntry::ParseResult()" );
					QLOGRAW( "Got a referral:" );
					QLOGRAW( "Message ID: %d", m_nMsgID );
					QLOGRAW( "URI: %s", full_ref_str );
				}
			else
				{
					QLOG( "QReadEntry::ParseResult()" );
					QLOGRAW( "ldap_url_parse failed:" );
					QLOGRAW( "Message ID: %d", m_nMsgID );
					QLOGRAW( "URI: %s",	referrals[0] );
				}

			if ( ludp != NULL )
				ldap_free_urldesc( ludp );

			free( ref_str );
		}
	else if ( rc != LDAP_SUCCESS )
    {
			//
			// Log the error
			QLOG( "QReadEntry::ParseResult()" );
			QLOGRAW( "An LDAP search resulted in error:" );
			QLOGRAW( "Message ID: %d", m_nMsgID );
			QLOGRAW( "LDAP error: %s", ldap_err2string( rc ) );

			//
			// Report the LDAP error!
			QReadErrorReport *pErrorReport = new QReadErrorReport();
			pErrorReport->Error() = QError( rc, pLDAPMsg );

			CString host, user, password;
			int port;
			m_DN.DS()->BindManager()->GetConnection
				( m_DN.GetConnID(), host, port, user, password );

			pErrorReport->DN() = m_DN;
			pErrorReport->Host() = host;

			Notify( DUAerror, pErrorReport );
    }
	else
		{
			//
			// Log the result
			QLOG( "QReadEntry::ParseResult()" );
			QLOGRAW( "LDAP search (read) returned a result:" );
			QLOGRAW( "Message ID: %d",	m_nMsgID );

			//
			// Report on the successful completion
			QReadCompletionReport *pReport = new QReadCompletionReport();
			CString host, user, password;
			int port;
			m_DN.DS()->BindManager()->GetConnection
				( m_DN.GetConnID(), host, port, user, password );

			pReport->DN() = m_DN;
			pReport->Host() = host;

			Notify( DUAcomplete, pReport );
		}

	QRequest::finished = TRUE;
	QRequest::invoked = FALSE;
	m_nMsgID = -1;
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::Abort
//
// Abort any outstanding LDAP opeations.
void QReadEntry::Abort()
{
	if ( m_nMsgID != -1 && invoked == TRUE && finished == FALSE )
		{
			ldap_abandon_ext( GetLDAPContext(), m_nMsgID, NULL, NULL );
		}
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::WaitingOn
//
// Find out if this request is waiting on an operation
// on the given connection.
BOOL QReadEntry::WaitingOn( ConnID connID )
{
	if ( connID == m_DN.GetConnID() && invoked == TRUE && finished == FALSE )
		return TRUE;

	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// QReadEntry::NotifyCloseConnection
//
// Notify of abnormal server connection closure
void QReadEntry::NotifyCloseConnection( ConnID conn_id )
{
	if ( WaitingOn( conn_id ) == FALSE )
		return;

	Abort();
	Notify( DUAabort, NULL );
}

