// QFormSearch.cpp: implementation of the QFormSearch class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Query32.h"
#include "QFormOp.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////

QFormOp::QFormOp
	( QDSManager *ds_arg,
		QForm *form_arg,
		ConnID conn_id_arg,
		QDN &base_dn_arg,
		CString &filter_arg )
{
	ds = ds_arg;
	form = form_arg;
	conn_id = conn_id_arg;
	base_dn = base_dn_arg;
	filter = filter_arg;
  op_invoked = op_done = op_succeeded = FALSE;
	ldap_msg_id = -1;
}

//////////////////////////////////////////////////////////////////////
// Destruction
//////////////////////////////////////////////////////////////////////

QFormOp::~QFormOp()
{
	WIN32_GUFF();

	POSITION pos = match_list.GetHeadPosition();
	while ( pos != NULL )
		{
			QEntry *entry = match_list.GetNext( pos );
			delete entry;
		}
}

//////////////////////////////////////////////////////////////////////
// Invoke a search operation using the given search specification
// (return LDAP message ID).
//////////////////////////////////////////////////////////////////////

BOOL QFormOp::Invoke( QError *&err )
{
	WIN32_GUFF();

	err = NULL;
	match_num = 0;

	//
	// Make sure connection is alive
	if ( ds->BindManager()->BindWait( conn_id, err ) == FALSE )
		return FALSE;

	//
	// Get search base DN in string form
	CString dn_str;
	base_dn.GetDN( dn_str );

	//
	// Compute search scope
	scope =
		( base_dn.GetRDNCount() >= 2 ? LDAP_SCOPE_SUBTREE : LDAP_SCOPE_ONELEVEL );

	//
	// XXX Attribute selection fixed at objectClass at present. Will need to
	// XXX be configurable later (QFormOp::Invoke).
	char *attrs[2];
	attrs[0] = "objectClass";
	attrs[1] = NULL;

	//
	// No timeout for the search operation
	struct timeval zerotime;
	zerotime.tv_sec = zerotime.tv_usec = 0L;
	
	//
	// Invoke LDAP search operation
	int ldap_err = ldap_search_ext
		( GetLDAPContext(),	const_cast<char *>( (LPCTSTR) dn_str ),	scope,
			const_cast<char *>( (LPCTSTR) filter ), NULL, 0,
			NULL, NULL, NULL, -1, &ldap_msg_id );

	op_invoked = TRUE;

	//
	// Check for errors
	if ( ldap_err != LDAP_SUCCESS )
		{
			CString err_str;
			err = new QError( ldap_err );
			err->Print( err_str );

			QLOG( "QFormOp::Invoke()" );
			QLOGRAW( "LDAP search invocation failed:" );
			QLOGRAW( "LDAP error: %s ", err_str );
			QLOGRAW
				( "DN: %s\nFilter: %s\nScope: %d\nMessage ID: %d",
					 dn_str, filter, scope, ldap_msg_id );

			ds->BindManager()->Dump( conn_id );
			return FALSE;
		}
	else
		{
			QLOG( "QFormOp::Invoke()" );
			QLOGRAW( "LDAP search invoked OK:" );
			QLOGRAW
				( "DN: %s\nFilter: %s\nScope: %d,\nMessage ID: %d",
					 dn_str, filter, scope, ldap_msg_id );

			ds->BindManager()->Dump( conn_id );
		}

	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Abort the operation in progress
//////////////////////////////////////////////////////////////////////

void QFormOp::Abort()
{
	if ( ldap_msg_id != -1 && op_invoked == TRUE && op_done == FALSE )
		ldap_abandon_ext( GetLDAPContext(), ldap_msg_id, NULL, NULL );

	op_done = TRUE;
	op_succeeded = FALSE;
}

//////////////////////////////////////////////////////////////////////
// Get matches. Matches may be returned asynchronously.
//////////////////////////////////////////////////////////////////////

UINT QFormOp::GetMatches( PtrList<QEntry> &match_list_arg )
{
	WIN32_GUFF();

	POSITION pos = match_list.GetHeadPosition();
	while ( pos != NULL )
		{
			match_list_arg.AddTail( match_list.GetNext( pos ) );
		}

	match_list.RemoveAll();
	return match_list_arg.GetCount();
}

//////////////////////////////////////////////////////////////////////
// Get returned referrals. Referrals may be returned asynchronously.
//////////////////////////////////////////////////////////////////////

UINT QFormOp::GetReferrals( QStringList &ref_list_arg )
{
	WIN32_GUFF();

	POSITION pos = referral_list.GetHeadPosition();
	while ( pos != NULL )
		{
			ref_list_arg.AddTail( referral_list.GetNext( pos ) );
		}

	return ref_list_arg.GetCount();
}

//////////////////////////////////////////////////////////////////////
// Handle an LDAP message 
//////////////////////////////////////////////////////////////////////

BOOL QFormOp::HandleLDAPMessage( LDAP *ld, LDAPMessage *lm, QError *& err )
{
	BOOL ok;
	switch ( ldap_msgtype( lm ) )
		{
		case LDAP_RES_SEARCH_ENTRY:
			ok = AddMatchingEntry( lm, err );
			break;

		case LDAP_RES_SEARCH_REFERENCE:
			ok = AddReferrals( lm ,err );
			break;

		case LDAP_RES_SEARCH_RESULT:
			ok = ParseResult( lm, err );
			break;
		}

	return ok;
}

//////////////////////////////////////////////////////////////////////
// Handle an LDAP message : LDAP_RES_SEARCH_ENTRY
//////////////////////////////////////////////////////////////////////

BOOL QFormOp::AddMatchingEntry( LDAPMessage *lm, QError *&err )
{
	int nCount = 0;
	for ( LDAPMessage *lme = ldap_first_entry( GetLDAPContext(), lm );
				lme != NULL;
				lme = ldap_next_entry( GetLDAPContext(), lm ) )
		{
			QEntry *entry = new QEntry( GetDS(), GetConnID(), GetLDAPContext(), lm );
			if ( entry->IsSet() == TRUE )
				{
					CString strDN;
					entry->DN().GetDN( strDN );

					QLOG( "QFormOp::AddMatchingEntry()" );
					QLOGRAW( "Got a match from an LDAP search:" );
					QLOGRAW( "Message ID: %d", ldap_msgid( lm ) );
					QLOGRAW( "DN: %s", strDN );

					nCount++;
					match_list.AddTail( entry );

					match_num++;
					return TRUE;
				}
			else
				{
					delete entry;
				}
		}

	return ( nCount > 0? TRUE : FALSE );
}

//////////////////////////////////////////////////////////////////////
// Handle an LDAP message : LDAP_RES_SEARCH_REFERENCE
//////////////////////////////////////////////////////////////////////

BOOL QFormOp::AddReferrals( LDAPMessage *lm, QError *&err )
{
	// XXX Unimplemented QFormOp::AddReferrals
	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// Handle an LDAP message  : LDAP_RES_SEARCH_RESULT
//////////////////////////////////////////////////////////////////////

BOOL QFormOp::ParseResult( LDAPMessage *lm, QError *&err )
{
  LDAPControl **serverctrls;
  char *matched_msg = NULL, *error_msg = NULL, **referrals;
	int ref_count;

  int rc;
  int parse_rc = ldap_parse_result
    ( GetLDAPContext(), lm, &rc, &matched_msg, &error_msg, &referrals, &serverctrls, 0 );

  if ( parse_rc != LDAP_SUCCESS )
    {
      // XXX Log LDAP failure from ldap_parse_result (QFormOp::ParseResult)
      // fprintf( stderr, "ldap_parse_result: %s\n", ldap_err2string( parse_rc ) );
      return FALSE;
    }

  // Check the results of the LDAP search operation.
	if ( rc == LDAP_REFERRAL )
		{
			op_succeeded = TRUE;
			for ( ref_count = 0; referrals[ref_count] != NULL; ref_count++ )
				{
					char *ref_str = strdup( referrals[ref_count] );
					LDAPURLDesc *ludp = NULL;

					int ref_ok = ldap_url_parse( ref_str, &ludp );
					if ( ref_ok == LDAP_SUCCESS )
						{
							CString full_ref_str;
							CString dn_str;
							CString scope_str;

							//
							// Get DN and seacrh scope string
							base_dn.GetDN( dn_str );
							scope_str = ( ludp->lud_scope == LDAP_SCOPE_SUBTREE ? "sub" : "one" );

							//
							// Format the referral string
							full_ref_str.Format
								( "ldap://%s:%d/%s?objectClass?%s?%s",
									ludp->lud_host,
									ludp->lud_port,
									ludp->lud_dn,
									( LPCTSTR ) scope_str,
									( LPCTSTR ) filter );

							referral_list.AddTail( full_ref_str );

							QLogFile::Log
								( "QFormOp::ParseResult()\
									\nGot a referral:\
									\nMessage ID: %d\
									\nURI: %s",
									ldap_msgid( lm ),
									full_ref_str );
						}
					else
						{
							QLogFile::Log
								( "QFormOp::ParseResult()\
									\nldap_url_parse failed:\
									\nMessage ID: %d\
									\nURI: %s",
									ldap_msgid( lm ),	referrals[0] );
						}

					if ( ludp != NULL )
						ldap_free_urldesc( ludp );

					free( ref_str );
				}
		}
	else if ( rc != LDAP_SUCCESS )
    {
			op_succeeded = FALSE;
			QLogFile::Log
				( "QFormOp::ParseResult()\
					\nAn LDAP search resulted in error:\
					\nLDAP error:%s",
					ldap_err2string( rc ) );

			Dump();

			//
			// If the error is due to an unavailable DSA close the connection
			if ( rc == LDAP_UNAVAILABLE || rc == LDAP_SERVER_DOWN )
				{
					ds->NotifyConnectionClose( conn_id );
				}
    }
	else
		{
			op_succeeded = TRUE;
			QLogFile::Log
				( "QFormOp::ParseResult()\
					\nLDAP search returned a result:\
					\nMatches: %d",
					match_num );

			Dump();
		}

	op_done = TRUE;
	return op_succeeded;
}

//////////////////////////////////////////////////////////////////////
// QFormOp::Dump
// Dump the form operation to the log file.
void QFormOp::Dump()
{
	CString host, user, password;
	int port;

	CString dn_str;
	base_dn.GetDN( dn_str );

	ds->BindManager()->GetConnection( conn_id, host, port, user, password );
	QLOGRAW
		( "Host: %s\nPort: %d\nDN: %s\nFilter: %s\nScope: %d\nMessage ID: %d",
			host, port, dn_str, filter, scope, ldap_msg_id );
}

